# SPDX-FileCopyrightText: 2024-present liuyanzhi08 <702368372@qq.com>
#
# SPDX-License-Identifier: MIT
# __version__ = "0.0.31" # 增加zmq支持
# __version__ = "0.0.32" # 抽取subscribe和publisher的原型
# __version__ = "0.0.33" # cal_exp增加safe_keys参数，传入以跳过'open'等关键词
# __version__ = "0.0.33" # fix zmq 的subscribe和publisher的context冲突
# __version__ = "0.0.34" # zmq支持是否server配置，允许1对多
# __version__ = "0.0.35" # get_trading_time默认返回原来交易所时区
# __version__ = "0.0.36" # get_trading_time默认返回原来交易所时区
# __version__ = "0.0.37" # 重构和修正cal_exp
# __version__ = "0.0.38" # cal_exp增加isin支持
# __version__ = "0.0.39" # cal_exp支持True+True -> 2
# __version__ = "0.0.41" # update version
# __version__ = "0.0.42" #修正cal_exp基本类型
# __version__ = "0.0.43" #cal_exp增加nanmin
__version__ = "0.0.44" # logger初始化时候就建立，而非第一次log才建立


from enum import Enum

class RetCode(Enum):
    OK = 'OK'
    ERROR = 'ERROR'
    ASYNC = 'ASYNC'

RET_OK = RetCode.OK
RET_ERROR = RetCode.ERROR
RET_ASYNC = RetCode.ASYNC

class Environment(Enum):
    REAL = 'REAL'
    TEST = 'TEST'


import zmq
import json
import threading
import time

class ZmqPublisher():
    # server表示是服务端，是创建端口的，客户端则是连接端口。服务端只有一个，客户端有多个
    def __init__(self, pub_address, server: bool = False):
        pub_context = zmq.Context()
        self.pub_context = pub_context

        #  Socket to talk to server
        self.publisher = pub_context.socket(zmq.PUB)
        if server:
            self.publisher.bind(pub_address)
        else:
            self.publisher.connect(pub_address)
        
    def send(self, topic: str, msg: dict):
        self.publisher.send_multipart([topic.encode('utf8'), json.dumps(msg).encode('utf8')])

    def stop(self):
        # 关闭 socket 和 context
        self.publisher.close()
        self.pub_context.term()

class ZmqSubscriber():
    def __init__(self, sub_address: str, server: bool = False):
        sub_context = zmq.Context()
        self.sub_context = sub_context
        self.running = True

        self.subscriber = sub_context.socket(zmq.SUB)
        
        if server:
            self.subscriber.bind(sub_address)
        else:
            self.subscriber.connect(sub_address)

        # topic = get_topic('.vntrader1', ZmqAction.SUB)
        self.subscriber.setsockopt_string(zmq.SUBSCRIBE, '')

        self.start_event = threading.Event() # 标记listen线程是否已经开始工作

        # 启动一个线程来运行 listen 方法
        self.listener_thread = threading.Thread(target=self.listen)
        self.listener_thread.start()
        self.start_event.wait()
        print('ZmqSubscriber inited') # todo!!

    def on_message(self, topic: str, msg: str):
        pass

    def listen(self):
        self.start_event.set()
        while self.running:
            try:
                topic_raw, msg_raw = self.subscriber.recv_multipart()
                # print(msg_raw)
                topic = topic_raw.decode()
                msg = msg_raw.decode()
                self.on_message(topic, msg)
            except zmq.Again as e:
                # 没有消息到达，可以继续循环
                # print(e)
                pass
            except zmq.ZMQError as e:
                # 处理 ZeroMQ 错误
                # breakpoint() # todo
                print(f"ZMQError: {e}")
                # break
            except Exception as e:
                # 处理其他异常
                # breakpoint() # todo
                print(f"Exception: {e}")
                # break     

    def send(self, topic: str, msg: dict):
        self.socket.send_multipart([topic.encode('utf8'), json.dumps(msg).encode('utf8')])

    def stop(self):
        self.running = False
        self.subscriber.close()
        self.sub_context.term()





from sympy import sympify, lambdify, Symbol, Number
from sympy.core.relational import Relational
from sympy.core.numbers import Float, Integer
from sympy.logic.boolalg import Boolean
from pandas import DataFrame
import pandas as pd
from typing import Union
import numpy as np
from .object import BaseException

BASIC_TYPES = [int, float, str, list, Float, Integer]

def is_basic_type(a):
    for basic_type in BASIC_TYPES:
        if isinstance(a, basic_type):
            return True
    return False

def convert(x):
    if isinstance(x, Number):
        val = float(x)
        return int(val) if val == int(val) else val
    return x

class MathsException(BaseException):
    pass

class ExpressionCache:
    """用于缓存表达式和对应矢量化函数的类"""
    _cache = {}

    @classmethod
    def get_calculator(cls, expression, safe_keys = []):
        """获取或创建表达式的矢量化计算器"""
        if expression not in cls._cache:
            cls._cache[expression] = VectorizedExpressionCalculator(expression, safe_keys)
        return cls._cache[expression]

# 支持 判断a字段在b字段的数组中(b字段是一个数组)
def isin_row(a, b):
    return [ai in bi for ai, bi in zip(a, b)]

def nanmean(*args):
    # args 是若干个 pandas Series（列）
    return np.array([np.nanmean(row) for row in zip(*args)])

def nansum(*args):
    # args 是若干个 pandas Series（列）
    return np.array([np.nansum(row) for row in zip(*args)])

def nanmax(*args):
    # args 是若干个 pandas Series（列）
    return np.array([np.nanmax(row) for row in zip(*args)])

def nanmin(*args):
    # args 是若干个 pandas Series（列）
    return np.array([np.nanmin(row) for row in zip(*args)])

class VectorizedExpressionCalculator:
    """矢量化表达式计算类"""
    def __init__(self, expression, safe_keys = []):
        # 将表达式字符串转换为 SymPy 表达式
        self.variables = []
        self.expression = sympify(expression, locals={k: Symbol(k) for k in safe_keys})
        # 将 SymPy 表达式转换为矢量化函数（支持 NumPy/Pandas 运算）
        
        # 如果是传入基本的数据类型，例如数组之类的，直接返回
        if not is_basic_type(self.expression):
            self.variables = sorted(self.expression.free_symbols, key=lambda x: str(x)) # 这里要用sorted很关键，因为set是没有顺序的，会导致值计算错误
            self.vectorized_func = lambdify(
                self.variables,
                self.expression,
                modules=[{
                    'max': np.maximum,
                    'min': np.minimum,
                    '&': np.logical_and,
                    '|': np.logical_or,
                    '==': np.equal,
                    '>': np.greater,
                    '<': np.less,
                    '>=': np.greater_equal,
                    '<=': np.less_equal,
                    'isin': isin_row,  # 覆盖原本的 np.isin
                    'nansum': nansum,
                    'nanmean': nanmean,
                    'nanmax': nanmax,
                    'nanmin': nanmin,
                }, 'numpy']
            )
        else:
            # 直接返回len(df) * x 个值
            self.calculate_dataframe = lambda df: [self.expression] * len(df)

    def calculate_dataframe(self, df):
        # 确保 DataFrame 中包含表达式需要的所有变量
        required_columns = [str(x) for x in self.variables]
        missing_columns = set(required_columns) - set(df.columns)
        if missing_columns:
            raise ValueError(f"Missing required columns: {missing_columns}")
        
        # 新增：将布尔列转为整数
        inputs = []
        for col in required_columns:
            s = df[col]
            if s.dtype == bool:
                s = s.astype(int)
            inputs.append(s)
        
        # 使用矢量化函数计算
        return self.vectorized_func(*inputs)


def calculate_expression(expression, df):
    """接口函数，直接传入表达式和 DataFrame 进行计算"""
    calculator = ExpressionCache.get_calculator(expression)
    return calculator.calculate_dataframe(df)

def cal_exp(exp: str, values: Union[dict, DataFrame], safe_keys: list[str]=[]):
    try:
        single_run = isinstance(values, dict)
        if single_run:
            values = pd.DataFrame([values])
        calculator = ExpressionCache.get_calculator(exp, safe_keys)
        series = pd.Series(calculator.calculate_dataframe(values))
        series = series.apply(convert)
        return series if not single_run else (series.iloc[0] if len(series) else None)
    except Exception as e:
        raise MathsException(f"表达式{exp}计算错误：{str(e)}", data={'values': values})

# t = cal_exp('a+b+c', pd.DataFrame({'a': [1,2,3], 'b': [2,3,4], 'c': [3,4,5]}))
# cal_exp('where(a>=2, where(b>=4, b, 999), c)', pd.DataFrame({'a': [1,2,3], 'b': [2,3,4], 'c': [3,4,5]}))
# print(t)


from decimal import Decimal
from math import floor, ceil

def to_decimal(value) -> Decimal:
    return Decimal(str(value))

def round_to(value: float, target: float) -> float:
    """
    Round price to price tick value.
    """
    value: Decimal = Decimal(str(value))
    target: Decimal = Decimal(str(target))
    rounded: Decimal = Decimal(str(int(round(value / target)) * target))
    return rounded


def floor_to(value: float, target: float) -> float:
    """
    Similar to math.floor function, but to target float number.
    """
    value: Decimal = Decimal(str(value))
    target: Decimal = Decimal(str(target))
    result: Decimal = Decimal(str(int(floor(value / target)) * target))
    return result


def ceil_to(value: float, target: float) -> float:
    """
    Similar to math.ceil function, but to target float number.
    """
    value: Decimal = Decimal(str(value))
    target: Decimal = Decimal(str(target))
    result: Decimal = Decimal(str(int(ceil(value / target)) * target))
    return result